#
# Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import threading
import json
import time
from threading import Lock
from netq_lib.common.netq_pq import QueueElem
from netq_lib.common.enums import TransMode, ConfKw
from netq_lib.orm import HOSTNAME
from netq_lib.common.enums import NetlinkMsgType
from netq_agent.interface.netq_opta_stream_pb2 import NetqAgentMessage


class GrpcClientThread(threading.Thread):
    def run(self):
        # Variable that stores the exception, if raised
        self.exc = None
        try:
            super(GrpcClientThread, self).run()
        except Exception as e:
            self.exc = str(e)

    def join(self, timeout=None):
        threading.Thread.join(self, timeout)
        # Since join() returns in caller thread
        # we re-raise the caught exception
        # if any was caught
        return self.exc


class GrpcClientBase(object):
    def __init__(self, conf, logger, pq, bp_pq, client, connection_callback=None):
        self._pq = pq
        self._bp_pq = bp_pq
        self._logger = logger
        self._channel = None
        self._active = False
        self._state_mu = Lock()
        self.handle_connection_cb = connection_callback
        self.seq_no = 0
        self.client = client
        self._conf = conf
        # GRPC CLIENT RETRY
        #self.__grpc_client_retry_default_status_codes = ["UNAVAILABLE", "DEADLINE_EXCEEDED"]
        self.__grpc_client_retry_default_status_codes = ["DEADLINE_EXCEEDED"]
        self.__grpc_client_retry_max_attempts = self._conf[ConfKw.NETQ_AGENT].get(ConfKw.GRPC_CLIENT_RETRY_MAX_ATTEMPTS, ConfKw.DEFAULT_GRPC_CLIENT_RETRY_MAX_ATTEMPTS)
        self.__grpc_client_retry_initial_backoff = self._conf[ConfKw.NETQ_AGENT].get(ConfKw.GRPC_CLIENT_RETRY_INITIAL_BACKOFF, ConfKw.DEFAULT_GRPC_CLIENT_RETRY_INITIAL_BACKOFF)
        self.__grpc_client_retry_max_backoff = self._conf[ConfKw.NETQ_AGENT].get(ConfKw.GRPC_CLIENT_RETRY_MAX_BACKOFF, ConfKw.DEFAULT_GRPC_CLIENT_RETRY_MAX_BACKOFF)
        self.__grpc_client_retry_backoff_multiplier = self._conf[ConfKw.NETQ_AGENT].get(ConfKw.GRPC_CLIENT_RETRY_BACKOFF_MULTIPLIER, ConfKw.DEFAULT_GRPC_CLIENT_RETRY_BACKOFF_MULTIPLIER)
        self._grpc_client_retry_enabled = self._conf[ConfKw.NETQ_AGENT].get(ConfKw.IS_GRPC_CLIENT_RETRY_ENABLED, ConfKw.DEFAULT_GRPC_CLIENT_RETRY_ENABLED)
        self._grpc_client_retry_status_codes = self.__grpc_client_retry_default_status_codes
        self._logger.info("GRPC client retry enabled: {}, max-attempts: {}, initial-backoff: {}, max-backoff: {}, backoff-multiplier: {}"
                          .format(self._grpc_client_retry_enabled, self.__grpc_client_retry_max_attempts, self.__grpc_client_retry_initial_backoff,
                                  self.__grpc_client_retry_max_backoff, self.__grpc_client_retry_backoff_multiplier))

    def get_grpc_channel_default_config(self):
        config_json = json.dumps(
            {
                "methodConfig": [
                    {
                        "name": [{}],
                        "retryPolicy": {
                            "maxAttempts": self.__grpc_client_retry_max_attempts,
                            "initialBackoff": self.__grpc_client_retry_initial_backoff,
                            "maxBackoff": self.__grpc_client_retry_max_backoff,
                            "backoffMultiplier": self.__grpc_client_retry_backoff_multiplier,
                            "retryableStatusCodes": self._grpc_client_retry_status_codes,
                        },
                    }
                ]
            }
        )
        return config_json

    def stop(self):
        self._state_mu.acquire()
        self._active = False
        self._state_mu.release()
        if self._channel:
            self._channel.close()
        self._channel = None

    def mark_active(self, state):
        self._state_mu.acquire()
        self._active = state
        self._state_mu.release()

    def is_active(self):
        return self._active

    def start_contiguous_stream(self):
        self.mark_active(True)
        try:
            self.run_contiguous_stream()
        except Exception as e:
            self._logger.error("start_contiguous_stream failed error {}".format(e))
            raise e
        finally:
            self.mark_active(False)

    def run_contiguous_stream(self):
        raise Exception("run_contiguous_stream method not implemented")

    def handle_connection(self):
        if not self.handle_connection_cb:
            raise Exception("handle_connection_cb is missing")
        if self._channel:
            self.stop()
        return self.handle_connection_cb()

    def get_sequence_number(self):
        self.seq_no = self.seq_no + 1
        return self.seq_no

    def send_pause_state(self, state):
        self._logger.info("Sending %s pause state %s", self.client, state)
        try:
            q_elem = QueueElem(0, self.client, state, self.client)
            self._bp_pq.putq(q_elem)
            if state is not True:
                for queue_idx in range(0, self._pq.max_num_queue):
                    self._pq.empty_pq(queue_idx)
            else:
                self._reset_grpc_backoff()
            self._logger.info("Agent %s Connectivity State %s",
                              self.client, state)

        except Exception as ex:
            self._logger.error('GRPC %s backpressure putq failed for %s',
                               self.client, ex)

    def _reset_grpc_backoff(self):
        self._logger.info("Resetting %s GRPC backoff time", self.client)
        self._grpc_backoff_time = 0

    def message_builder(self, msg_type, fullstate, json_out):
        full_state = TransMode.FULL_TRANSFORM if fullstate else TransMode.PARTIAL_TRANSFORM

        if isinstance(json_out, list):
            if len(json_out) > 0:
                event_time = int(json_out[0].get('timestamp', 0) * 1000)
            else:
                event_time = int(time.time() * 1000)
            output = json.dumps(json_out)
        elif isinstance(json_out, str):
            event_time = int(time.time() * 1000)
            output = json_out
        else:
            event_time = int(time.time() * 1000)
            output = json.dumps([json_out])

        netq_message = NetqAgentMessage(aid=HOSTNAME,
                                        message_type=msg_type,
                                        seqno=self.get_sequence_number(),
                                        ts=event_time,
                                        trans_mode=full_state,
                                        type=0,
                                        message=output.encode('utf-8'))
        return netq_message
